
/* module.c: Sony TLC59116 Driver.
 *
 * Copyright 2016 Sony Corporation.
 * Author: Sony Corporation.
 *
 * This package is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * THIS PACKAGE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/mutex.h>
#include <linux/workqueue.h>
#include <linux/i2c.h>
#include <linux/leds.h>
#include <linux/dma-mapping.h>
#include <linux/earlysuspend.h>
#include <mach/mt_gpio.h>
#include "tlc59116.h"


#define DRV_NAME					"tlc59116"
//#define DELAYWRITE

#if 0
	#define DBGTRACE(fmt, ...)		printk("\e[32m[" DRV_NAME "]       " fmt "\e[0m\n", ## __VA_ARGS__)
	#define ERRTRACE(fmt, ...)		printk("\e[31m[" DRV_NAME "]       " fmt "\e[0m\n", ## __VA_ARGS__)
#else
	#define DBGTRACE(fmt, ...)		//
	#define ERRTRACE(fmt, ...)		printk(      "[" DRV_NAME "] error " fmt      "\n", ## __VA_ARGS__)
#endif


	enum {
		I2C_SPEED					=	400,						/* I2C 通信速度     */
		DEVICE_MAX					=	16,							/* デバイス最大数   */
		PWM_MAX						=	16,							/* PWM 最大数       */
		DELAYTIME					=	20							/* 書き込み遅延時間 */
	} ;

	struct priv ;

	/* デバイス情報 */
	struct deviceinfo {
		u8							init ;							/* 初期化フラグ */
		u8							addr ;							/* I2C アドレス */
		u8							reserved[2] ;					/* 未使用       */
		u8							ledout[4];						/* LEDOUT 値    */
		u8							value[16] ;						/* 値           */
		u8							value_old[16] ;					/* 以前の値     */
	} ;

	/* PWM LED 情報 */
	struct pwminfo {
		struct led_classdev			cdev ;							/* LED デバイスクラス */
		struct priv					*p_priv ;						/* プライベートデータ */
		u16							count ;							/* PWM データ数 x 3   */
		u8							flag ;							/* フラグ             */
		u8							reserved ;						/* 未使用             */
		u8							const *p_pwm_data ;				/* PWM データ         */
	} ;

	/* プライベートデータ */
	struct priv {
		struct i2c_client			*p_client ;						/* I2C クライアント           */
		struct tlc59116_config		*p_config ;						/* 構成                       */
		struct mutex				mutex ;							/* ミューテックス             */
		struct workqueue_struct		*p_workqueue ;					/* ワークキュー               */
		struct delayed_work			work ;							/* ワーク                     */
		struct early_suspend		early_suspend ;					/* サスペンド                 */
		uint						buff_size ;						/* 送信バッファサイズ         */
		dma_addr_t					buff_addr ;						/* 送信バッファの物理アドレス */
		u8							*p_buff ;						/* 送信バッファ               */
		u8							suspend ;						/* サスペンドフラグ           */
		u8							device_count ;					/* デバイス数                 */
		u8							pwm_count ;						/* PWM LED 数                 */
		struct pwminfo				*p_pwminfo ;					/* PWM LED 情報               */
		struct deviceinfo			deviceinfo[1] ;					/* デバイス情報               */
	} ;


/* -------------------------------------------------------------------------------- */
/*   メッセージを作成                                                               */
/*                                                                                  */
/*   IN : p_priv = プライベートデータ                                               */
/*        p_msg  = メッセージ格納アドレス                                           */
/*   OUT: メッセージ数                                                              */
/* -------------------------------------------------------------------------------- */
static int create_msg(struct priv * p_priv, struct i2c_msg *p_msg)
{
	struct deviceinfo	*p_deviceinfo,  *p_deviceinfo_end ;
	int					count, length ;
	uint				head, tail, size ;

	/* 初期化 */
	p_deviceinfo     = &p_priv->deviceinfo[0] ;
	p_deviceinfo_end = &p_priv->deviceinfo[p_priv->device_count] ;
	size             = 0 ;
	count            = 0 ;

	do {
		if( p_deviceinfo->init ) {

			/* 変更範囲取得 */
			head   = 0 ;
			tail   = PWM_MAX - 1 ;
			length = 0 ;

			do {
				/* 先頭 */
				if( p_deviceinfo->value[head] != p_deviceinfo->value_old[head] ) {
					do {
						if( p_deviceinfo->value[tail] != p_deviceinfo->value_old[tail] ) {
							break ;
						}
					} while( head < --tail ) ;
					length = tail - head + 1 ;
					break ;
				}

				/* 最後 */
				if( p_deviceinfo->value[tail] != p_deviceinfo->value_old[tail] ) {
					while( ++head < tail ) {
						if( p_deviceinfo->value[head] != p_deviceinfo->value_old[head] ) {
							break ;
						}
					}
					length = tail - head + 1 ;
					break ;
				}
			} while( ++head < --tail ) ;

			if( length ) {
#if 0
uint	index ;

printk("\e[32m[" "0x%02x - %d, %d, %d : ", p_deviceinfo->addr, head, tail, length) ;
for(index=0; index<length; index++) {
	printk("%02X ", p_deviceinfo->value[head+index]) ;
}
printk("\e[0m\n") ;
#endif
				/* 以前の値へ保存 */
				memcpy(&p_deviceinfo->value_old[head], &p_deviceinfo->value[head], length) ;

				/* メッセージ作成 */
				memcpy(&p_priv->p_buff[size+1], &p_deviceinfo->value[head], length) ;
				p_priv->p_buff[size] = 0x82 + head ;
				p_msg->addr          = p_deviceinfo->addr ;
				p_msg->flags         = I2C_M_STOP ;
				p_msg->len           = 1 + length ;
				p_msg->buf           = (u8 *)(p_priv->buff_addr + size) ;
				p_msg->timing        = I2C_SPEED ;
				p_msg->ext_flag      = I2C_DMA_FLAG ;

				size += p_msg->len ;
				p_msg++ ;
				count++ ;
			}
		}
	} while( ++p_deviceinfo != p_deviceinfo_end ) ;

	return count ;
}


/* -------------------------------------------------------------------------- */
/*   ワークキュー処理                                                         */
/*                                                                            */
/*   IN : p_wok = プライベートデータ内のワークアドレス                        */
/*   OUT: なし                                                                */
/* -------------------------------------------------------------------------- */
static void workqueue_proc( struct work_struct *p_work )
{
	struct priv		*p_priv=container_of(p_work, struct priv, work.work) ;
	struct i2c_msg	msg[DEVICE_MAX] ;
	int				result ;
	uint			count ;

	/* メッセージ作成 */
	mutex_lock( &p_priv->mutex ) ;

	count = create_msg(p_priv, &msg[0]) ;

	mutex_unlock( &p_priv->mutex ) ;

	if( count ) {
		result = i2c_transfer(p_priv->p_client->adapter, &msg[0], count) ;
		if( result != count ) {
			ERRTRACE("%s(%d) : %d", __func__, __LINE__, result) ;
		}
	}
}


/* -------------------------------------------------------------------------------- */
/*   PWM の値を設定                                                                 */
/*                                                                                  */
/*   IN : p_priv   = プライベートデータ                                             */
/*        pwm_data = PWM データ                                                     */
/*        value    = 値                                                             */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static inline void set_pwm_value(struct priv *p_priv, u8 pwm_data, u8 value)
{
	if( pwm_data != TLC59116__________ ) {
		p_priv->deviceinfo[pwm_data>>4].value[pwm_data&15] = value ;
	}
}


/* -------------------------------------------------------------------------------- */
/*   明るさを設定（PWM）                                                            */
/*                                                                                  */
/*   IN : p_cdev = LED デバイスクラス                                               */
/*        value  = 明るさ                                                           */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void set_brightness_pwm(struct led_classdev *p_cdev, enum led_brightness value)
{
	struct pwminfo	*p_pwminfo=container_of(p_cdev, struct pwminfo, cdev) ;
	struct priv		*p_priv=p_pwminfo->p_priv ;
	u8				const *p_pwm_data, *p_pwm_data_end ;


	if( p_priv->suspend ) {
		return ;
	}

	if( p_pwminfo->count ) {
		mutex_lock( &p_priv->mutex ) ;

		p_pwm_data     = &p_pwminfo->p_pwm_data[0] ;
		p_pwm_data_end = &p_pwminfo->p_pwm_data[p_pwminfo->count] ;

		do {
			set_pwm_value(p_priv, *p_pwm_data++, (u8 )(value>>16)) ;	/* R */
			set_pwm_value(p_priv, *p_pwm_data++, (u8 )(value>> 8)) ;	/* G */
			set_pwm_value(p_priv, *p_pwm_data++, (u8 )(value>> 0)) ;	/* B */
		} while( p_pwm_data != p_pwm_data_end ) ;

		mutex_unlock( &p_priv->mutex ) ;
	}

#ifndef DELAYWRITE
	if( p_pwminfo->flag & TLC59116_SYNC ) {
		queue_delayed_work(p_priv->p_workqueue, &p_priv->work, 0) ;
	}
#else
	cancel_delayed_work( &p_priv->work ) ;

	if( p_pwminfo->flag & TLC59116_SYNC ) {
		queue_delayed_work(p_priv->p_workqueue, &p_priv->work, 0) ;
	} else {
		queue_delayed_work(p_priv->p_workqueue, &p_priv->work, msecs_to_jiffies( DELAYTIME )) ;
	}
#endif/*DELAYWRITE*/
}


/* -------------------------------------------------------------------------------- */
/*   PWM LED 情報へ設定                                                             */
/*                                                                                  */
/*   IN : p_priv    = プライベートデータ                                            */
/*        p_pwminfo = PWM LED 情報                                                  */
/*        p_pwm     = PWM LED                                                       */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void set_pwminfo(struct priv *p_priv, struct pwminfo *p_pwminfo, struct tlc59116_pwm const *p_pwm)
{
	struct deviceinfo	*p_deviceinfo ;
	u8					const *p_pwm_data, *p_pwm_data_end ;
	u8					pwm_data ;
	uint				pwm_num ;

	/* 初期化 */
	p_pwminfo->cdev.name           = &p_pwm->p_name[0] ;
	p_pwminfo->cdev.brightness     = 0x00000000 ;
	p_pwminfo->cdev.max_brightness = 0x00ffffff ;
	p_pwminfo->cdev.brightness_set = set_brightness_pwm ;
	p_pwminfo->p_priv              = p_priv ;
	p_pwminfo->count               = 3 * p_pwm->count ;
	p_pwminfo->flag                = p_pwm->flag ;
	p_pwminfo->p_pwm_data          = &p_priv->p_config->p_pwm_data[3*p_pwm->index] ;

	/* LEDOUT 設定 */
	p_pwm_data     = &p_pwminfo->p_pwm_data[0] ;
	p_pwm_data_end = &p_pwminfo->p_pwm_data[p_pwminfo->count] ;

	while( p_pwm_data != p_pwm_data_end ) {
		pwm_data = *p_pwm_data++ ;
		if( pwm_data != TLC59116__________ ) {
			p_deviceinfo                     = &p_priv->deviceinfo[pwm_data>>4] ;
			pwm_num                          = pwm_data & 15 ;
			p_deviceinfo->ledout[pwm_num/4] |= (2 << ((pwm_num & 3) * 2)) ;
		}
	}
}


/* -------------------------------------------------------------------------------- */
/*   リセット                                                                       */
/*                                                                                  */
/*   IN : p_priv = プライベートデータ                                               */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void reset( struct priv *p_priv )
{
	static const u8		mode[]={ 0x00, 0x00 } ;
	struct i2c_msg		msg[2] ;
	struct deviceinfo	*p_deviceinfo, *p_deviceinfo_end ;
	u8					ledout[1+4] ;
	int					result ;
	uint				index ;

	/* リセット */
	mt_set_gpio_out(p_priv->p_config->reset, GPIO_OUT_ZERO) ;
	msleep( 10 ) ;

	mt_set_gpio_out(p_priv->p_config->reset, GPIO_OUT_ONE) ;
	msleep( 10 ) ;

	/* 初期化 */
	for(index=0; index<p_priv->pwm_count; index++) {
		p_priv->p_pwminfo[index].cdev.brightness = 0 ;
	}

	memset(&msg[0], 0, sizeof( msg )) ;
	msg[0].flags     = I2C_M_STOP ;			/* MODE1 */
	msg[0].len       = sizeof( mode ) ;
	msg[0].buf       = (u8 *)&mode[0] ;
	msg[0].timing    = I2C_SPEED ;
	msg[1].flags     = I2C_M_STOP ;			/* LEDOUT0-3 */
	msg[1].len       = sizeof( ledout ) ;
	msg[1].timing    = I2C_SPEED ;
	msg[1].buf       = (u8 *)&ledout[0] ;
	ledout[0]        = 0x94 ;
	p_deviceinfo     = &p_priv->deviceinfo[0] ;
	p_deviceinfo_end = &p_priv->deviceinfo[p_priv->device_count] ;

	do {
		memset(&p_deviceinfo->value[0],     0, sizeof( p_deviceinfo->value )) ;
		memset(&p_deviceinfo->value_old[0], 0, sizeof( p_deviceinfo->value_old )) ;
		p_deviceinfo->init = 0 ;

		memcpy(&ledout[1], &p_deviceinfo->ledout[0], sizeof( p_deviceinfo->ledout )) ;
		msg[0].addr = p_deviceinfo->addr ;
		msg[1].addr = p_deviceinfo->addr ;
		result      = i2c_transfer(p_priv->p_client->adapter, &msg[0], ARRAY_SIZE( msg )) ;
		if( result == ARRAY_SIZE( msg ) ) {
			p_deviceinfo->init = 1 ;
		}
	} while( ++p_deviceinfo != p_deviceinfo_end ) ;
}


/* -------------------------------------------------------------------------------- */
/*   サスペンド処理                                                                 */
/*                                                                                  */
/*   IN : p_early_suspend = サスペンド                                              */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void tlc59116_early_suspend( struct early_suspend *p_early_suspend )
{
	struct priv		*p_priv=container_of(p_early_suspend, struct priv, early_suspend) ;


	DBGTRACE("%s(%d)", __func__, __LINE__) ;

	p_priv->suspend = 1 ;

	cancel_delayed_work( &p_priv->work ) ;
	flush_workqueue( p_priv->p_workqueue ) ;

	/* リセット状態 */
	mt_set_gpio_out(p_priv->p_config->reset, GPIO_OUT_ZERO) ;
}


/* -------------------------------------------------------------------------------- */
/*   レジューム処理                                                                 */
/*                                                                                  */
/*   IN : p_early_suspend = サスペンド                                              */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void tlc59116_early_resume( struct early_suspend *p_early_suspend )
{
	struct priv		*p_priv=container_of(p_early_suspend, struct priv, early_suspend) ;


	DBGTRACE("%s(%d)", __func__, __LINE__) ;

	/* リセット */
	reset( p_priv ) ;

	p_priv->suspend = 0 ;
}


/* -------------------------------------------------------------------------------- */
/*   プローブ                                                                       */
/*                                                                                  */
/*   IN : p_client = I2C クライアント                                               */
/*        p_id     = I2C デバイスの識別子                                           */
/*   OUT: 0       == 正常終了                                                       */
/*        0       <  異常終了                                                       */
/* -------------------------------------------------------------------------------- */
static int tlc59116_probe(struct i2c_client *p_client, const struct i2c_device_id *p_id)
{
	struct tlc59116_config	*p_config=p_client->dev.platform_data ;
	struct priv				*p_priv ;
	struct led_classdev		*p_cdev ;
	int						result, index ;
	uint					size ;

	/* プライベートデータ確保 */
	size   = offsetof(struct priv, deviceinfo) + (sizeof( struct deviceinfo ) * p_config->device_count) +
	                                             (sizeof( struct pwminfo )    * p_config->pwm_count) ;
	p_priv = kzalloc(size, GFP_KERNEL) ;
	if( p_priv == NULL ) {
		ERRTRACE("%s(%d)", __func__, __LINE__) ;
		return -ENOMEM ;
	}

	/* 初期化 */
	i2c_set_clientdata(p_client, p_priv) ;
	mutex_init( &p_priv->mutex ) ;
	p_priv->p_client     = p_client ;
	p_priv->p_config     = p_config ;
	p_priv->device_count = p_config->device_count ;
	p_priv->pwm_count    = p_config->pwm_count ;
	p_priv->p_pwminfo    = (struct pwminfo  *)&p_priv->deviceinfo[p_priv->device_count] ;
	result               = 0 ;

	for(index=0; index<p_priv->device_count; index++) {
		p_priv->deviceinfo[index].addr = p_config->p_addr[index] ;
	}

	for(index=0; index<p_priv->pwm_count; index++) {
		set_pwminfo(p_priv, &p_priv->p_pwminfo[index], &p_config->p_pwm[index]) ;
	}

	/* DMA バッファ確保 */
	p_priv->buff_size = (1 + PWM_MAX) * p_priv->device_count ;
	p_priv->p_buff    = dma_alloc_coherent(0, p_priv->buff_size, &p_priv->buff_addr, GFP_KERNEL) ;
	if( p_priv->p_buff == NULL ) {
		ERRTRACE("%s(%d)", __func__, __LINE__) ;
		result = -ENOMEM ;
		goto tlc59116_probe_f00 ;
	}

	/* ワークキュー作成 */
	p_priv->p_workqueue = create_singlethread_workqueue( DRV_NAME ) ;
	if( p_priv->p_workqueue == NULL ) {
		ERRTRACE("%s(%d)", __func__, __LINE__) ;
		result = -ENOMEM ;
		goto tlc59116_probe_e00 ;
	}
	INIT_DELAYED_WORK(&p_priv->work, workqueue_proc) ;

	/* LED クラスデバイス登録 */
	for(index=0; index<p_priv->pwm_count; index++) {
		p_cdev = &p_priv->p_pwminfo[index].cdev ;
		result = led_classdev_register(NULL, p_cdev) ;
		if( result < 0 ) {
			ERRTRACE("%s(%d)", __func__, __LINE__) ;
			for(--index; index>=0; index--) {
				led_classdev_unregister( &p_priv->p_pwminfo[index].cdev ) ;
			}
			goto tlc59116_probe_d00 ;
		}
		result = sysfs_chmod_file(&p_cdev->dev->kobj, &p_cdev->dev->class->dev_attrs->attr, 0666) ;
	}

	/* リセット */
	reset( p_priv ) ;

	/* サスペンド登録 */
	p_priv->early_suspend.level   = EARLY_SUSPEND_LEVEL_BLANK_SCREEN + 1 ;
	p_priv->early_suspend.suspend = tlc59116_early_suspend ;
	p_priv->early_suspend.resume  = tlc59116_early_resume ;
	register_early_suspend( &p_priv->early_suspend ) ;

	return 0 ;


tlc59116_probe_c00:
	for(index=0; index<p_priv->pwm_count; index++) {
		led_classdev_unregister( &p_priv->p_pwminfo[index].cdev ) ;
	}
tlc59116_probe_d00:
	destroy_workqueue( p_priv->p_workqueue ) ;
tlc59116_probe_e00:
	dma_free_coherent(0, p_priv->buff_size, &p_priv->p_buff[0], p_priv->buff_addr) ;
tlc59116_probe_f00:
	kfree( p_priv ) ;

	return result ;
}


/* -------------------------------------------------------------------------------- */
/*   削除                                                                           */
/*                                                                                  */
/*   IN : p_client = I2C クライアント                                               */
/*   OUT: 常に０                                                                    */
/* -------------------------------------------------------------------------------- */
static int tlc59116_remove( struct i2c_client *p_client )
{
	struct priv		*p_priv=i2c_get_clientdata( p_client ) ;
	uint			index ;


	unregister_early_suspend( &p_priv->early_suspend ) ;

	for(index=0; index<p_priv->pwm_count; index++) {
		led_classdev_unregister( &p_priv->p_pwminfo[index].cdev ) ;
	}

	cancel_delayed_work( &p_priv->work ) ;
	flush_workqueue( p_priv->p_workqueue ) ;
	destroy_workqueue( p_priv->p_workqueue ) ;

	/* リセット状態 */
	mt_set_gpio_out(p_priv->p_config->reset, GPIO_OUT_ZERO) ;

	dma_free_coherent(0, p_priv->buff_size, &p_priv->p_buff[0], p_priv->buff_addr) ;

	kfree( p_priv ) ;

	return 0 ;
}


static const struct i2c_device_id	tlc59116_id[] = {
										{ DRV_NAME, 0 },
										{}
									} ;

static struct i2c_driver			tlc59116_driver = {
										.probe     = tlc59116_probe,
										.remove    = tlc59116_remove,
										.id_table  = tlc59116_id,
										.driver    = {
											.name  = DRV_NAME,
											.owner = THIS_MODULE,
										},
									} ;


module_i2c_driver( tlc59116_driver ) ;


MODULE_DESCRIPTION("Sony TLC59116 Driver");
MODULE_AUTHOR("Sony Corporation");
MODULE_LICENSE("GPL") ;
MODULE_ALIAS("Sony TLC59116 Driver") ;
MODULE_VERSION( "v1.01" ) ;
